GtkScrollbar#xviewer-image-gallery-scrollbar {
	-GtkScrollbar-has-backward-stepper: false;
	-GtkScrollbar-has-secondary-backward-stepper: false;
	-GtkScrollbar-has-forward-stepper: false;
	-GtkScrollbar-has-secondary-forward-stepper: false;
	-GtkScrollbar-min-slider-length: 60;
}
